'''
Implementation of the Spaghetti model training using PyTorch Lightning
'''
import os
import torch
import torch.nn as nn
import pytorch_lightning as pl
from pytorch_lightning.loggers import CSVLogger
from torchvision.utils import save_image
import itertools
from . import _spaghetti_modules as sp_modules
from . import utils

class _Spaghetti(pl.LightningModule):
    '''
    Implement the training steps for the SPAGHETTI model
    args:
        save_dir: str, the directory to save the model checkpoints and logs
        batch_size: int, the batch size for the model, default 1
        weights: list of floats, the weights for the loss functions in the order of 
        GAN loss, cycle loss, identity loss, and SSIM loss. Default [1.0, 10.0, 5.0, 10.0]
        lr: float, the learning rate for the model, default 0.0002
        
    '''
    def __init__(self, save_dir, batch_size = 1, weights = [1.0, 10.0, 5.0, 10.0],
                 lr = 0.0005):
        super().__init__()
        self.save_dir = save_dir
        self.automatic_optimization = False
        # model
        self.G_AB = sp_modules.GeneratorResNet(3, 9)
        self.D_B = sp_modules.Discriminator(3)
        self.G_BA = sp_modules.GeneratorResNet(3, 9)
        self.D_A = sp_modules.Discriminator(3)
        # loss
        self.criterion_GAN = nn.MSELoss()
        self.criterion_cycle = nn.L1Loss()
        self.criterion_identity = nn.L1Loss()
        self.criterion_ssim = sp_modules.SSIMLoss()
        self.weights = weights
        # others
        self.batch_size = batch_size
        self.lr = lr

    def calculate_loss_generator(self, res, x1, x2):
        # groud truth
        out_shape = [x1.size(0), 1, x1.size(2)//self.D_A.scale_factor, 
                     x1.size(3)//self.D_A.scale_factor]
        valid = torch.ones(out_shape).to(self.device)
        fake_x1, fake_x2, recov_x1, recov_x2, new_x1, new_x2 = res
        # convert to gray scale for ssim calculation
        gray = Grayscale(num_output_channels=3)
        x1_gray = gray(x1)
        x2_gray = gray(x2)
        fake_x1_gray = gray(fake_x1)
        fake_x2_gray = gray(fake_x2)
        new_x1_gray = gray(new_x1)
        new_x2_gray = gray(new_x2)
        loss_GAN = (self.criterion_GAN(self.D_A(fake_x1), valid) 
                    + self.criterion_GAN(self.D_B(fake_x2), valid))/2
        loss_cycle = (self.criterion_cycle(recov_x1, x1) 
                      + self.criterion_cycle(recov_x2, x2))/2
        loss_identity = (self.criterion_identity(new_x1, x1) 
                         + self.criterion_identity(new_x2, x2))/2
        loss_ssim_fake = (self.criterion_ssim(fake_x2_gray, x1_gray) 
                     + self.criterion_ssim(fake_x1_gray, x2_gray)) / 2
        loss_ssim_real = (self.criterion_ssim(new_x1_gray, x1_gray)
                        + self.criterion_ssim(new_x2_gray, x2_gray)) / 2
        loss_ssim = (loss_ssim_fake + loss_ssim_real) / 2
        total_loss = (self.weights[0] * loss_GAN + self.weights[1] * loss_cycle 
                      + self.weights[2] * loss_identity + self.weights[3] * loss_ssim)
        return total_loss

    def calculate_loss_discriminator(self, dis, x, x_fake):
        out_shape = [x.size(0), 1, x.size(2)//dis.scale_factor, 
                     x.size(3)//dis.scale_factor]
        valid = torch.ones(out_shape).to(self.device)
        fake = torch.zeros(out_shape).to(self.device)
        loss_real = self.criterion_GAN(dis(x), valid)
        loss_fake = self.criterion_GAN(dis(x_fake.detach()), fake)
        total_loss = (loss_real + loss_fake) /2
        return total_loss
    
    def training_step(self, batch, batch_idx):
        optimizer_G, optimizer_D_A, optimizer_D_B = self.optimizers()
        accumulated_grad_batches = (batch_idx+1) % self.batch_size == 0
        x1, x2 = batch
        # for discriminator
        fake_x1 = self.G_BA(x2)
        fake_x2 = self.G_AB(x1)
        
        # for identity
        new_x1 = self.G_BA(x1)
        new_x2 = self.G_AB(x2)

        # for reconstruction
        recov_x2 = self.G_AB(fake_x1)
        recov_x1 = self.G_BA(fake_x2)
        
        res = (fake_x1, fake_x2, recov_x1, recov_x2, new_x1, new_x2)
        
        # generator loss
        gen_loss = self.calculate_loss_generator(res, x1, x2)

        # discriminator A loss
        d_a_loss = self.calculate_loss_discriminator(self.D_A, x1, fake_x1)
        
        # discriminator B loss
        d_b_loss = self.calculate_loss_discriminator(self.D_B, x2, fake_x2)

        # optmize
        total_loss = gen_loss + d_a_loss + d_b_loss
        self.manual_backward(total_loss)
        
        if accumulated_grad_batches:
            optimizer_G.step()
            optimizer_G.zero_grad()
            optimizer_D_B.step()
            optimizer_D_B.zero_grad()
            optimizer_D_A.step()
            optimizer_D_A.zero_grad()

        metrics = {'train_gen_loss': gen_loss, 'train_D_A_loss': d_a_loss, 'train_D_B_loss': d_b_loss}
        self.log_dict(metrics,prog_bar=True)

    def validation_step(self, batch, batch_idx):
        x1, x2 = batch
        fake_x1 = self.G_BA(x2)
        fake_x2 = self.G_AB(x1)
        
        # for identity
        new_x1 = self.G_BA(x1)
        new_x2 = self.G_AB(x2)

        # for reconstruction
        recov_x2 = self.G_AB(fake_x1)
        recov_x1 = self.G_BA(fake_x2)
        
        res = (fake_x1, fake_x2, recov_x1, recov_x2, new_x1, new_x2)

        loss = self.calculate_loss_generator(res, x1, x2)
        self.log_dict({'val_gen_loss': loss},sync_dist=True,prog_bar=True)
        # save image for visualization
        visual = torch.cat((x1, fake_x2, x2, fake_x1), 0)
        if batch_idx % 1000 == 0:
            try:
                save_image(visual, os.path.join(self.save_dir, "visual", 
                                                f"visual_rank_{str(self.global_rank)}_batch_{batch_idx}_epoch_{self.current_epoch}.png"), 
                        nrow=4, normalize=True, value_range=(-1, 1))
            except FileExistsError:
                pass 

    def configure_optimizers(self):
        optimizer_G = torch.optim.AdamW(
            itertools.chain(self.G_AB.parameters(), self.G_BA.parameters()), 
            lr=self.lr, weight_decay=1e-2)
        optimizer_D_A = torch.optim.AdamW(
            self.D_A.parameters(), 
            lr=self.lr, weight_decay=1e-2)
        optimizer_D_B = torch.optim.AdamW(
            self.D_B.parameters(), 
            lr=self.lr, weight_decay=1e-2)
        return [optimizer_G, optimizer_D_A, optimizer_D_B], []

def train_spaghetti(train_loader, val_loader, batch_size=1, weights=[1.0, 10.0, 5.0, 10.0],
                    lr = 0.0002, save_dir = None, epochs=100, name="my_spaghetti",
                    num_nodes=1, ngpus_per_node="auto"):
    '''
    Train the SPAGHETTI model using PyTorch Lightning.
    args:
        train_loader: the PyTorch Dataloader for the training dataset
        val_loader: the PyTorch Dataloader for the validation dataset
        batch_size: int, the batch size for the model, default 1
        weights: list of floats, the weights for the loss functions in the order of 
        GAN loss, cycle loss, identity loss, and SSIM loss. Default [1.0, 10.0, 5.0, 10.0]
        lr: float, the learning rate for the model, default 0.0002
        save_dir: str, the directory to save the model checkpoints and logs. Default current directory
        epochs: int, the number of epochs to train the model, default 100
        name: str, the name of the model for the logger, default "my_spaghetti"
        num_nodes: int, the number of nodes to train the model, default 1
        ngpus_per_node: int, the number of GPUs per node, default "auto" to use all the available GPUs
    '''
    if save_dir is None:
        final_save_dir = os.getcwd()
    else:
        final_save_dir = save_dir
    if not os.path.exists(os.path.join(final_save_dir, f"{name}_visualization")):
        os.makedirs(os.path.join(final_save_dir, f"{name}_visualization"))
    # create model
    lit_model = _Spaghetti(save_dir=final_save_dir, batch_size=batch_size, weights=weights, lr=lr)
    # train model
    logger = CSVLogger(final_save_dir, name=name)
    trainer = pl.Trainer(max_epochs=epochs, devices=ngpus_per_node, num_nodes=num_nodes,
                         use_distributed_sampler=True, enable_progress_bar=True,
                         strategy="ddp",
                         default_root_dir=final_save_dir, logger=logger)
    print("Trainer initialized with ", ngpus_per_node, "GPU(s) per node on ", num_nodes, "node(s)")
    print("Training Starting...")
    ckpt = utils.find_checkpoint(final_save_dir)
    if ckpt:
        print("Checkpoint found. Resuming from ", ckpt)
    else:
        print("Starting from epoch 0")
    trainer.fit(lit_model, train_loader, val_loader, None, ckpt)
    print("Training ended.")


'''
Modules for the SPAGHETTI model to translate microscopy images to H&E images
'''
import torch.nn as nn
import torch.nn.functional as F
import torch

class ResidualBlock(nn.Module):
    def __init__(self, in_channels):
        super(ResidualBlock, self).__init__()
        self.block = nn.Sequential(
            nn.ReflectionPad2d(1), # padding, keep the image size constant after next conv2d
            nn.Conv2d(in_channels, in_channels, 3),
            nn.InstanceNorm2d(in_channels),
            nn.ReLU(inplace=True),
            nn.ReflectionPad2d(1),
            nn.Conv2d(in_channels, in_channels, 3),
            nn.InstanceNorm2d(in_channels)
        )
    
    def forward(self, x):
        return x + self.block(x)
    
class GeneratorResNet(nn.Module):
    def __init__(self, in_channels, num_residual_blocks=9):
        super(GeneratorResNet, self).__init__()
        
        # Inital Convolution  3*256*256 -> 64*256*256
        out_channels=64
        self.conv = nn.Sequential(
            nn.ReflectionPad2d(in_channels), # padding, keep the image size constant after next conv2d
            nn.Conv2d(in_channels, out_channels, 2*in_channels+1),
            nn.InstanceNorm2d(out_channels),
            nn.ReLU(inplace=True),
        )
        
        channels = out_channels
        
        # Downsampling   64*256*256 -> 128*128*128 -> 256*64*64
        self.down = []
        for _ in range(2):
            out_channels = channels * 2
            self.down += [
                nn.Conv2d(channels, out_channels, 3, stride=2, padding=1),
                nn.InstanceNorm2d(out_channels),
                nn.ReLU(inplace=True),
            ]
            channels = out_channels
        self.down = nn.Sequential(*self.down)
        
        # Transformation (ResNet)  256*64*64
        self.trans = [ResidualBlock(channels) for _ in range(num_residual_blocks)]
        self.trans = nn.Sequential(*self.trans)
        
        # Upsampling  256*64*64 -> 128*128*128 -> 64*256*256
        self.up = []
        for _ in range(2):
            out_channels = channels // 2
            self.up += [
                nn.Upsample(scale_factor=2), # bilinear interpolation
                nn.Conv2d(channels, out_channels, 3, stride=1, padding=1),
                nn.InstanceNorm2d(out_channels),
                nn.ReLU(inplace=True),
            ]
            channels = out_channels
        self.up = nn.Sequential(*self.up)
        
        # Out layer  64*256*256 -> 3*256*256
        self.out = nn.Sequential(
            nn.ReflectionPad2d(in_channels),
            nn.Conv2d(channels, in_channels, 2*in_channels+1),
            nn.Tanh()
        )
    
    def forward(self, x):
        x = self.conv(x)
        x = self.down(x)
        x = self.trans(x)
        x = self.up(x)
        x = self.out(x)
        return x

class Discriminator(nn.Module):
    def __init__(self, in_channels):
        super(Discriminator, self).__init__()
        
        self.model = nn.Sequential(
            *self.block(in_channels, 64, normalize=False), # 3*256*256 -> 64*128*128 
            *self.block(64, 128),  # 64*128*128 -> 128*64*64
            *self.block(128, 256), # 128*64*64 -> 256*32*32
            *self.block(256, 512), # 256*32*32 -> 512*16*16
            
            nn.ZeroPad2d((1,0,1,0)), # padding left and top   512*16*16 -> 512*17*17
            nn.Conv2d(512, 1, 4, padding=1) # 512*17*17 -> 1*16*16
        )
        
        self.scale_factor = 16
    
    @staticmethod
    def block(in_channels, out_channels, normalize=True):
        layers = [nn.Conv2d(in_channels, out_channels, 4, stride=2, padding=1)]
        if normalize:
            layers.append(nn.InstanceNorm2d(out_channels))
        layers.append(nn.LeakyReLU(0.2, inplace=True))
        
        return layers
        
    def forward(self, x):
        return self.model(x)

class SSIMLoss(nn.Module):
    def __init__(self, window_size=11, size_average=True):
        super(SSIMLoss, self).__init__()
        self.window_size = window_size
        self.size_average = size_average
        self.channel = 1
        self.create_window(window_size)
        
    def create_window(self, window_size, channel=1):
        # Create a Gaussian window (filter) with specified size and channel
        sigma = 1.5
        exponents = torch.Tensor([-(x - window_size//2)**2 / float(2 * sigma**2) for x in range(window_size)])
        gauss = torch.exp(exponents)
        gauss = gauss / gauss.sum()
        _1D_window = gauss.unsqueeze(1)
        _2D_window = _1D_window.mm(_1D_window.t()).float().unsqueeze(0).unsqueeze(0)
        window = _2D_window.expand(channel, 1, window_size, window_size).contiguous()
        self.window = window

    def forward(self, img1, img2):
        if img1.size(1) != self.channel:
            self.channel = img1.size(1)
            self.create_window(self.window_size, self.channel)
        
        # Move window to the same device as the images
        window = self.window.to(img1.device)
        
        # Compute SSIM components
        mu1 = F.conv2d(img1, window, padding=self.window_size//2, groups=self.channel)
        mu2 = F.conv2d(img2, window, padding=self.window_size//2, groups=self.channel)
        
        mu1_sq = mu1.pow(2)
        mu2_sq = mu2.pow(2)
        mu1_mu2 = mu1 * mu2

        sigma1_sq = F.conv2d(img1 * img1, window, padding=self.window_size//2, groups=self.channel) - mu1_sq
        sigma2_sq = F.conv2d(img2 * img2, window, padding=self.window_size//2, groups=self.channel) - mu2_sq
        sigma12 = F.conv2d(img1 * img2, window, padding=self.window_size//2, groups=self.channel) - mu1_mu2

        # Set SSIM constants
        C1 = 0.01 ** 2
        C2 = 0.03 ** 2

        # SSIM calculation
        ssim_map = ((2 * mu1_mu2 + C1) * (2 * sigma12 + C2)) / ((mu1_sq + mu2_sq + C1) * (sigma1_sq + sigma2_sq + C2))
        
        if self.size_average:
            return torch.clamp((1 - ssim_map.mean()) / 2, 0, 1)  # SSIM loss as (1 - SSIM) / 2
        else:
            return torch.clamp((1 - ssim_map) / 2, 0, 1)  # If per-sample loss required



'''
Useful functions for SPAGHETTI
'''

import os

def find_checkpoint(dir: str):
    '''
    Find the latest checkpoint in the directory
    args:
        dir: str, the directory to search for the checkpoint for Pytorch Lightning
    return:
        str or None. The str of the path to the latest checkpoint that ends with .ckpt
        If no checkpoint is found, return None
    '''
    files = []
    for path, _, file in os.walk(dir):
        for f in file:
            if f.endswith(".ckpt"):
                files.append(os.path.join(path, f))
    if len(files) == 0:
        return None
    else:
        return max(files, key=os.path.getctime)


'''
Perform pre-processing on the images using the pre-trained model
'''
import torch
from spaghetti import _spaghetti_modules as sp_modules
import os
from tqdm import tqdm
from torchvision.utils import save_image
import torchvision.transforms.v2 as v2

class Spaghetti():
    '''
    The class to perform the inference using the SPAGHETTI model
    '''
    def __init__(self, model_path: str):
        '''
        Initialize the model
        args:
            model_path: str, the path to the model checkpoint
        '''
        device = torch.device('cuda' if torch.cuda.is_available() else 'cpu')
        generator = sp_modules.GeneratorResNet(3, 9)
        generator.to(device)
        ckpt = torch.load(model_path, map_location=device)["state_dict"]
        # get only G_AB weights
        ckpt = {k[5:]: v for k, v in ckpt.items() if ("G_AB" in k)}
        generator.load_state_dict(ckpt)
        generator.eval()
        # set attributes
        self.generator = generator
        self.device = device

    @staticmethod
    def _default_transform(img):
        '''
        The default transformation to perform on the image
        args:
            img: np.ndarray or PIL.Image or torch.Tensor, the image to perform the transformation
        return:
            torch.Tensor, the transformed image
        '''
        transform = [v2.ToImage(),
                v2.ToDtype(torch.float32, scale=True),
                v2.Normalize(mean=[0.5, 0.5, 0.5], std=[0.5, 0.5, 0.5]),
                v2.Resize((256, 256)),
                ]
        return v2.Compose(transform)(img)

    def pre_processing(self, imgs, transform: None):
        '''
        Perform the pre-processing on the images
        args:
            imgs: list[PIL.Image or torch.Tensor or numpy.ndarray], the images to perform the pre-processing
            transform: None, "default", or callable, the transformation to perform on the images. 
            If None, no transformation is performed. If "default", the default transformation is performed.
        return:
            list[torch.Tensor], the images after the pre-processing
        '''
        processed_imgs = []
        for img in imgs:
            if transform == "default":
                transformed = self._default_transform(img)
            elif transform:
                transformed = transform(img)
            else:
                transformed = img
            processed_imgs.append(transformed)
        return processed_imgs

    def inference(self, imgs: list[torch.Tensor], names: list[str], save_path: None):
        '''
        Perform the inference on the image
        args:
            img: list[torch.Tensor], the image(s) to perform the inference
            names: list[str], the names of the images
            save_path: str or None. If str, images will be saved to the path to after the transformation
        return:
            list[torch.Tensor], the images after the SPAGHETTI transformaton
        '''
        print("Performing Inference...")
        transformed_imgs = []
        if save_path:
            if not os.path.exists(save_path):
                os.makedirs(save_path)
        with torch.no_grad():
            for idx, img in enumerate(tqdm(imgs)):
                img = img.unsqueeze(0).to(self.device)
                transformed = self.generator(img).squeeze(0).cpu()
                # normalize to range [0,1]
                out = torch.clamp(transformed, min=-1, max=1)
                min_val = out.min()
                max_val = out.max()
                out = (out-min_val)/(max(max_val-min_val, 1e-5))
                out = torch.clamp(out, min=0, max=1) # ensure no overflow
                if save_path:
                    name = os.path.join(save_path, f"transformed_{names[idx]}.png")
                    save_image(out, name, nrow=1, normalize=True, value_range=(-1, 1))
                transformed_imgs.append(out)
        return transformed_imgs


'''
Prepare the datasets for training and inference
'''
import random
from PIL import Image
import os
from torch.utils.data import Dataset

class TrainingDataset(Dataset):
    '''
    The dataset class for the SPAGHETTI model training
    args:
        path_1: list of strings, the paths to the images in domain 1
        path_2: list of strings, the paths to the images in domain 2
        transform_1: the transformation for domain 1 images
        transform_2: the transformation for domain 2 images
        num_sample: int, optional, the number of images to sample from each domain
    '''
    def __init__(self, path_1: list[str], path_2: list[str], 
                 transform_1, transform_2, num_sample=None):
        random.seed(42)
        # domain 1
        domain1_paths = []
        for each in path_1:
            domain1_paths.extend([os.path.join(each, x) for x in os.listdir(each) 
                                  if x.endswith((".png", ".jpg", ".jpeg", ".tiff", ".tif"))])
        if num_sample:
            try:
                self.domain1_images = random.sample(domain1_paths, k=num_sample)
            except ValueError:
                self.domain1_images = random.choices(domain1_paths, k=num_sample)
        else:
            self.domain1_images = domain1_paths
        
        # domain 2
        domain2_paths = []
        for each in path_2:
            domain2_paths.extend([os.path.join(each, x) for x in os.listdir(each)
                                  if x.endswith((".png", ".jpg", ".jpeg", ".tiff", ".tif"))])
        if num_sample:
            try:
                self.domain2_images = random.sample(domain2_paths, k=num_sample)
            except ValueError:
                self.domain2_images = random.choices(domain2_paths, k=num_sample)
        else:
            self.domain2_images = domain2_paths

        # others
        self.length_dataset = max(len(self.domain1_images), len(self.domain2_images))
        self.domain1_len = len(self.domain1_images)
        self.domain2_len = len(self.domain2_images)
        self.transform_1 = transform_1
        self.transform_2 = transform_2
        print("Dataset created with: ", self.length_dataset, " samples")

    def __len__(self):
        return self.length_dataset

    def __getitem__(self, index):
        domain1_img_path = self.domain1_images[index % self.domain1_len]
        domain2_img_path = self.domain2_images[index % self.domain2_len]
        domain1_img = Image.open(domain1_img_path).convert("RGB")
        domain2_img = Image.open(domain2_img_path).convert("RGB") 

        domain1_img = self.transform_1(domain1_img)
        domain2_img = self.transform_2(domain2_img)

        return domain1_img, domain2_img


from setuptools import setup, find_packages

setup(
    name="pcm-spaghetti",  
    version="1.1.1",  
    author="Richard (Zhi Fei) Dong, Chris McIntosh, Gregory W. Schwartz",
    author_email="gregory.schwartz@uhn.ca",
    description="A PyTorch implementation of the SPAGHETTI model for phase-contrast microscopy image transformation",
    long_description=open("README.md").read(),
    long_description_content_type="text/markdown",
    url="https://github.com/schwartzlab-methods/spaghetti",  
    packages=find_packages(exclude=["tutorials*"]),  
    entry_points={
        "console_scripts": [
            "spaghetti=spaghetti.cli_inference:main",
        ],
    },
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Operating System :: OS Independent",
    ],
    python_requires=">=3.9", 
    install_requires=[
        "torch>=2.2.0",
        "torchvision>=0.17.0",
        "numpy>=1.26.4",
        "Pillow>=10.3.0",
        "scikit-image>=0.18.3",
        "pytorch-lightning>=2.3.3",
        "tqdm>=4.66.4",
    ],
)



'''
Entry point for the CLI inference of SPAGHETTI.
'''
import os
import argparse
from spaghetti import inferences
from PIL import Image

def inference(input, output, checkpoint):
    '''
    The inference function for the CLI inference
    args:
        input: str, the input image or directory to translate
        output: str, the output directory to save the translated image(s)
        checkpoint: str, the path to the checkpoint file
    '''
    # check if input is a directory
    if os.path.isdir(input):
        # get all images
        imgs = []
        names = []
        for path, _, files in os.walk(input):
            for f in files:
                if f.endswith((".jpg", ".jpeg", ".png", ".tiff", ".tif")):
                    imgs.append(os.path.join(path, f))
                    names.append(f.split(".")[0])
    else:
        imgs = [input]
        names = [str(os.path.basename(input)).split(".")[0]]
    # create the model
    model = inferences.Spaghetti(checkpoint)
    # perform the inference
    pil_imgs = [Image.open(img).convert("RGB") for img in imgs]
    processed_imgs = model.pre_processing(pil_imgs, transform="default")
    model.inference(processed_imgs, names, output)

def main():
    parser = argparse.ArgumentParser(description="CLI for translating PCM images using SPAGHETTI")
    parser.add_argument("--input", '-i', type=str, help="The input image or directory to translate")
    parser.add_argument("--output", '-o', type=str, help="The output directory to save the translated image(s)")
    parser.add_argument("--checkpoint", '-c', type=str, help="The path to the checkpoint file", required=True)
    args = parser.parse_args()
    assert os.path.exists(args.checkpoint), "The checkpoint file does not exist"
    print("Welcome to SPAGHETTI CLI Inference")
    print("Checkpoint file found at ", args.checkpoint, "will be used for inference")
    print("Starting Inference...")
    inference(args.input, args.output, args.checkpoint)
    print("Inference Completed. Images saved to ", args.output)

if __name__ == "__main__":
    main()